#drinc:exec/miscellaneous.g
#drinc:exec/memory.g
#drinc:exec/nodes.g
#drinc:exec/ports.g
#drinc:exec/tasks.g
#drinc:exec/io.g
#drinc:exec/interrupts.g
#drinc:devices/input.g
#drinc:devices/inputevent.g
#drinc:devices/console.g
#drinc:devices/conunit.g
#drinc:graphics/gfx.g
#drinc:graphics/gfxbase.g
#drinc:graphics/rastport.g
#drinc:graphics/view.g
#drinc:graphics/copper.g
#drinc:libraries/dos.g
#drinc:disassemble.g

#ddb.g

[2] uint COLOUR_TABLE = (0x666, 0x000);

uint
    TEXT_COLUMNS = 80,			/* width in characters */
    OUTPUT_MAX = 50,			/* size of output buffer (< 80) */
    RAW_MAX = 256;			/* size of unread input buffer */

[RAW_MAX] char RawBuffer;		/* buffer of keyboard characters */
uint RawAddPoint, RawReadPoint; 	/* queue pointers */
uint RawCount;				/* count of characters in queue */
*Task_t MyTask;
ulong EventSignal;
bool EventWaiting, Suspended;

uint Width, Height;
*GfxBase_t GfxBase;
*View_t OldView;
View_t View;
BitMap_t BitMap;
ViewPort_t ViewPort;
RasInfo_t RasInfo;
RastPort_t RastPort;

[OUTPUT_MAX] char TextBuffer;
uint OutputPos, TextColumn;

*char InputBuffer, UserPrompt;		/* caller's input buffer & prompt */
uint InputPos;				/* count of input char's typed */
bool SpecialMode, CursorPresent, InInput;

/*
 * showCursor - show the text cursor.
 */

proc showCursor()void:

    RectFill(&RastPort, TextColumn * 8, Height - 8,
	     TextColumn * 8 + 7, Height - 1);
    CursorPresent := true;
corp;

/*
 * removeCursor - remove the text cursor.
 */

proc removeCursor()void:

    SetAPen(&RastPort, 0);
    RectFill(&RastPort, TextColumn * 8, Height - 8,
	     TextColumn * 8 + 7, Height - 1);
    SetAPen(&RastPort, 1);
    CursorPresent := false;
corp;

/*
 * scroll - scroll the output screen up one line.
 */

proc scroll()void:

    ScrollRaster(&RastPort, 0, 8, 0, 0, Width - 1, Height - 1);
    Move(&RastPort, 0, Height - 2);
corp;

/*
 * flushLine - flush the buffered output to the screen.
 */

proc flushLine()void:
    uint lenBefore, lenAfter;

    if OutputPos ~= 0 then
	if TextColumn + OutputPos >= TEXT_COLUMNS then
	    /* the output will wrap */
	    lenBefore := TEXT_COLUMNS - TextColumn;
	    if lenBefore ~= 0 then
		Text(&RastPort, &TextBuffer[0], lenBefore);
	    fi;
	    scroll();
	    lenAfter := OutputPos - lenBefore;
	    if lenAfter ~= 0 then
		Text(&RastPort, &TextBuffer[lenBefore], lenAfter);
	    fi;
	    TextColumn := lenAfter;
	else
	    Text(&RastPort, &TextBuffer[0], OutputPos);
	    TextColumn := TextColumn + OutputPos;
	fi;
	OutputPos := 0;
    fi;
corp;

/*
 * putChar - output a character to the buffered output line.
 */

proc putChar(char ch)void:

    if OutputPos = OUTPUT_MAX then
	flushLine();
    fi;
    if ch = '\n' then
	flushLine();
	scroll();
	TextColumn := 0;
    else
	TextBuffer[OutputPos] := ch;
	OutputPos := OutputPos + 1;
    fi;
corp;

/*
 * putStr - output a string to the buffered output line.
 */

proc putStr(register *char p)void:

    while p* ~= '\e' do
	putChar(p*);
	p := p + sizeof(char);
    od;
corp;

/*
 * backup - untype an input character.
 */

proc backup()void:

    if TextColumn = 0 then
	TextColumn := TEXT_COLUMNS - 1;
	ScrollRaster(&RastPort, 0, -8, 0, 0, Width - 1, Height - 1);
    else
	TextColumn := TextColumn - 1;
    fi;
    Move(&RastPort, TextColumn * 8, Height - 2);
corp;

/*
 * readLine - read in a new input command line.
 */

proc readLine(*char prompt, buffer; uint length)void:
    char ch;

    UserPrompt := prompt;
    InputBuffer := buffer;
    InInput := true;
    putStr(prompt);
    flushLine();
    showCursor();
    InputPos := 0;
    while
	Forbid();
	if RawCount = 0 then
	    EventWaiting := true;
	    Permit();
	    ignore Wait(EventSignal);
	    while SpecialMode do
		Delay(1);
	    od;
	    Forbid();
	fi;
	RawCount := RawCount - 1;
	Permit();
	ch := RawBuffer[RawReadPoint];
	RawReadPoint := RawReadPoint + 1;
	if RawReadPoint = RAW_MAX then
	    RawReadPoint := 0;
	fi;
	ch ~= '\r'
    do
	if ch = '\b' then
	    if InputPos ~= 0 then
		InputPos := InputPos - 1;
		removeCursor();
		backup();
		showCursor();
	    fi;
	elif ch = '\(0x18)' then
	    removeCursor();
	    while InputPos ~= 0 do
		InputPos := InputPos - 1;
		backup();
	    od;
	    showCursor();
	else
	    if InputPos < length then
		(buffer + InputPos * sizeof(char))* := ch;
		InputPos := InputPos + 1;
		removeCursor();
		putChar(ch);
		flushLine();
		showCursor();
	    fi;
	fi;
    od;
    (buffer + InputPos * sizeof(char))* := '\e';
    removeCursor();
    scroll();
    TextColumn := 0;
    InInput := false;
corp;

/*
 * startSpecial - output from another process - clean the input line.
 */

proc startSpecial()void:

    SpecialMode := true;
    if InInput then
	SetAPen(&RastPort, 0);
	RectFill(&RastPort, 0, Height - 8, Width - 1, Height - 1);
	SetAPen(&RastPort, 1);
    fi;
corp;

/*
 * proc endSpecial - end of special output mode.
 */

proc endSpecial()void:
    register uint i;
    register *char p;

    if InInput then
	putStr(UserPrompt);
	p := InputBuffer;
	i := InputPos;
	while i ~= 0 do
	    i := i - 1;
	    putChar(p*);
	    p := p + sizeof(char);
	od;
	flushLine();
	if CursorPresent then
	    showCursor();
	fi;
    fi;
    SpecialMode := false;
corp;

/*
 * setSuspend - allow upper level to control whether or not we are suspended.
 */

proc setSuspend(bool wantSuspend)void:

    if wantSuspend and not Suspended then
	Suspended := true;
	LoadView(OldView);
    elif not wantSuspend and Suspended then
	Suspended := false;
	OldView := GfxBase*.gb_ActiView;
	LoadView(&View);
    fi;
corp;

/*
 * inputHandler -
 *	Magic routine that is called by Exec when the input device has
 *	something for me.
 */

proc inputHandler()*InputEvent_t:
    uint
	BUFF_MAX = 10,

	OP_MOVEL = 0x2000,
	M_ADIR = 0o1,
	M_DISP = 0o5,
	R_FP = 6,
	A0 = 0,

	D_CODE = 0x22,	/* keyboard code for 'd' */
	ALL_LEFT =
	    IEQUALIFIER_LSHIFT | IEQUALIFIER_LALT | IEQUALIFIER_LCOMMAND;
    register **InputEvent_t prevP;
    register *InputEvent_t ie;
    *InputEvent_t head;
    register ushort theClass;
    register uint count;
    register *char p;
    [BUFF_MAX] char rawBuff;

    /* move A0 into 'head' */
    code(
	OP_MOVEL | R_FP << 9 | M_DISP << 6 | M_ADIR << 3 | A0,
	head
    );
    prevP := &head;
    ie := head;
    while ie ~= nil do
	theClass := ie*.ie_Class;
	if Suspended then
	    if theClass = IECLASS_RAWKEY and ie*.ie_Code = D_CODE and
		ie*.ie_Qualifier & ALL_LEFT = ALL_LEFT
	    then
		/* wake up on LSHIFT-LALT-LCOMMAND-d */
		Suspended := false;
		OldView := GfxBase*.gb_ActiView;
		LoadView(&View);
		/* we have consumed the event, do not pass it through */
		/* Hmmm. I wonder if anybody further down the chain will get
		   confused by only half of a keystroke? */
		ie := ie*.ie_NextEvent;
		prevP* := ie;
	    else
		/* otherwise, just pass it through */
		prevP := &ie*.ie_NextEvent;
		ie := ie*.ie_NextEvent;
	    fi;
	else
	    if theClass = IECLASS_RAWKEY then
		/* process the character and remove the event */
		if ie*.ie_Code & IECODE_UP_PREFIX = 0 then
		    count := RawKeyConvert(ie, &rawBuff[0], BUFF_MAX, nil);
		    if count > 0 and count < RAW_MAX - RawCount then
			Forbid();
			p := &rawBuff[0];
			RawCount := RawCount + count;
			while count ~= 0 do
			    count := count - 1;
			    RawBuffer[RawAddPoint] := p*;
			    RawAddPoint := RawAddPoint + 1;
			    if RawAddPoint = RAW_MAX then
				RawAddPoint := 0;
			    fi;
			    p := p + sizeof(char);
			od;
			if EventWaiting then
			    EventWaiting := false;
			    Signal(MyTask, EventSignal);
			fi;
			Permit();
		    fi;
		fi;
		/* don't pass this on to Intuition */
		ie := ie*.ie_NextEvent;
		prevP* := ie;
	    elif theClass = IECLASS_TIMER or
		theClass = IECLASS_POINTERPOS or
		theClass = IECLASS_DISKREMOVED or
		theClass = IECLASS_DISKINSERTED
	    then
		/* pass these on to Intuition */
		prevP := &ie*.ie_NextEvent;
		ie := ie*.ie_NextEvent;
	    else
		/* and chuck anything else - we don't want Intuition to be
		   responding to someone playing with the mouse and maybe
		   accidentally doing something nasty - like what happens if
		   they get hold of a screen drag bar and start dragging? */
		ie := ie*.ie_NextEvent;
		prevP* := ie;
	    fi;
	fi;
    od;
    head
corp;

/*
 * inputStuff - create and set up our input handler, then call command handler.
 */

proc inputStuff()void:
    *MsgPort_t inputDevPort;
    *IOStdReq_t inputStd;
    IOStdReq_t consoleStd;
    Interrupt_t myHandler;
    register *char p;
    register ushort tries;

    p := "ddb inputX";
    tries := 0;
    while
	(p + 9 * sizeof(char))* := tries + '0';
	FindPort(p) ~= nil
    do
	tries := tries + 1;
    od;
    MyTask := FindTask(nil);
    EventSignal := AllocSignal(-1);
    if EventSignal ~= -1 then
	inputDevPort := CreatePort(p, 0);
	if inputDevPort ~= nil then
	    inputStd := CreateStdIO(inputDevPort);
	    if inputStd ~= nil then
		if OpenDevice("input.device", 0, &inputStd*.io_io, 0) = 0 then
		    if OpenDevice("console.device", -1,
				  &consoleStd.io_io, 0) = 0
		    then
			SetConsoleDevice(consoleStd.io_io.io_Device);
			RawAddPoint := 0;
			RawReadPoint := 0;
			RawCount := 0;
			EventWaiting := false;
			Suspended := false;
			SpecialMode := false;
			CursorPresent := false;
			InInput := false;
			myHandler.is_Code := pretend(inputHandler, proc()void);
			myHandler.is_Node.ln_Pri := 75 - tries;
			inputStd*.io_io.io_Command := IND_ADDHANDLER;
			inputStd*.io_Data := pretend(&myHandler, *byte);
			ignore DoIO(&inputStd*.io_io);

			childInit();

			inputStd*.io_io.io_Command := IND_REMHANDLER;
			ignore DoIO(&inputStd*.io_io);
			CloseDevice(&consoleStd.io_io);
		    fi;
		    CloseDevice(&inputStd*.io_io);
		fi;
		DeleteStdIO(inputStd);
	    fi;
	    DeletePort(inputDevPort);
	fi;
	FreeSignal(EventSignal);
    fi;
corp;

/*
 * graphics - create and set up our graphics screen, then call input stuff
 */

proc graphics()void:

    Width := 640;
    Height := 400;
    InitView(&View);
    View.v_ViewPort := &ViewPort;
    View.v_Modes := LACE;
    InitBitMap(&BitMap, 1, Width, Height);
    BitMap.bm_Planes[0] := AllocRaster(Width, Height);
    if BitMap.bm_Planes[0] ~= nil then
	RasInfo.ri_BitMap := &BitMap;
	RasInfo.ri_RxOffset := 0;
	RasInfo.ri_RyOffset := 0;
	RasInfo.ri_Next := nil;
	InitVPort(&ViewPort);
	ViewPort.vp_Modes := HIRES | LACE;
	ViewPort.vp_DWidth := Width;
	ViewPort.vp_DHeight := Height;
	ViewPort.vp_RasInfo := &RasInfo;
	ViewPort.vp_ColorMap := GetColorMap(2);
	if ViewPort.vp_ColorMap ~= nil then
	    InitRastPort(&RastPort);
	    RastPort.rp_BitMap := &BitMap;
	    LoadRGB4(&ViewPort, &COLOUR_TABLE[0], 2);
	    SetRast(&RastPort, 0);
	    SetAPen(&RastPort, 1);
	    SetBPen(&RastPort, 0);
	    MakeVPort(&View, &ViewPort);
	    MrgCop(&View);
	    OldView := GfxBase*.gb_ActiView;
	    LoadView(&View);

	    OutputPos := 0;
	    TextColumn := 0;
	    scroll();
	    inputStuff();

	    LoadView(OldView);
	    FreeVPortCopLists(&ViewPort);
	    FreeCprList(View.v_LOFCprList);
	    FreeCprList(View.v_SHFCprList);
	    FreeColorMap(ViewPort.vp_ColorMap);
	fi;
	FreeRaster(BitMap.bm_Planes[0], Width, Height);
    fi;
corp;

/*
 * main - open libraries, scan command line parameters
 */

proc main()void:

    if OpenExecLibrary(0) ~= nil then
	if OpenDosLibrary(0) ~= nil then
	    GfxBase := OpenGraphicsLibrary(0);
	    if GfxBase ~= nil then
		if OpenDisassembleLibrary(0) ~= nil then
		    graphics();
		    CloseDisassembleLibrary();
		fi;
		CloseGraphicsLibrary();
	    fi;
	    CloseDosLibrary();
	fi;
	CloseExecLibrary();
    fi;
corp;
