#drinc:exec/miscellaneous.g
#drinc:exec/nodes.g
#drinc:exec/ports.g
#drinc:exec/tasks.g
#drinc:exec/memory.g
#drinc:libraries/dos.g
#drinc:libraries/dosextens.g

#ddb.g

*char PORT_NAME = "ddb portX";

type
    StartSeg_t = struct {
	ulong ss_length;
	ulong ss_nextSeg;
	uint ss_jsrCode;
	ulong ss_jsrAddr;
	uint ss_jmpCode;
	ulong ss_jmpAddr;
    };

*char PortName;
*StartSeg_t StartSeg;
*MsgPort_t ParentPort;
*Message_t TheMessage;
Segment_t ChildSegment;
*MsgPort_t ChildPort;
*Process_t ChildProcess;
ulong OldReturnAddress;
bool ParentWaiting, ChildHasRun;

/*
 * childStartup - routine inserted into the beginning of the child
 */

proc childStartup()void:
    register *MsgPort_t replyPort;

    replyPort := CreatePort(nil, 0);
    if replyPort ~= nil then
	TheMessage*.mn_Node.ln_Type := NT_MESSAGE;
	TheMessage*.mn_ReplyPort := replyPort;
	PutMsg(ParentPort, TheMessage);
	ignore WaitPort(replyPort);
	ignore GetMsg(replyPort);
	DeletePort(replyPort);
    else
	Delay(1000000000);
    fi;
corp;

/*
 * childTerminate - routine inserted into the child's termination.
 */

proc childTerminate()void:
    uint
	OP_MOVEL = 0x2000,
	OP_JMP = 0x4ec0,
	M_ADIR = 1,
	M_INDIR = 2,
	M_SPECIAL = 0o7,
	M_ABS_LONG = 0o1,
	A0 = 0;

    if ParentWaiting then
	TheMessage*.mn_ReplyPort := nil;
	PutMsg(ParentPort, TheMessage);
    fi;
    childDone();
    code(
	OP_MOVEL | A0 << 9 | M_ADIR << 6 | M_SPECIAL << 3 | M_ABS_LONG,
	OldReturnAddress,
	OP_JMP | M_INDIR << 3 | A0
    );
corp;

/*
 * childInit - initialize for child processing.
 */

proc childInit()void:
    uint
	OP_JMP = 0x4ec0,
	OP_JSR = 0x4e80,
	M_SPECIAL = 0o7,
	M_ABS_LONG = 0o1;
    uint tries;

    StartSeg := AllocMem(sizeof(StartSeg_t), MEMF_PUBLIC);
    if StartSeg ~= nil then
	StartSeg*.ss_length := sizeof(StartSeg_t);
	StartSeg*.ss_jsrCode := OP_JSR | M_SPECIAL << 3 | M_ABS_LONG;
	StartSeg*.ss_jsrAddr := pretend(childStartup, ulong);
	StartSeg*.ss_jmpCode := OP_JMP | M_SPECIAL << 3 | M_ABS_LONG;
	PortName := PORT_NAME;
	tries := 0;
	while
	    (PortName + 8)* := 'a' + tries;
	    FindPort(PortName) ~= nil
	do
	    tries := tries + 1;
	od;
	ParentPort := CreatePort(PortName, 0);
	if ParentPort ~= nil then
	    TheMessage :=
		AllocMem(sizeof(Message_t), MEMF_PUBLIC | MEMF_CLEAR);
	    if TheMessage ~= nil then
		ParentWaiting := false;
		ChildHasRun := false;

		commands();

		FreeMem(TheMessage, sizeof(Message_t));
	    fi;
	    DeletePort(ParentPort);
	fi;
	FreeMem(StartSeg, sizeof(StartSeg_t));
    fi;
corp;

/*
 * childLoad - load the child program.
 */

proc childLoad(*char childName)bool:

    ChildSegment := LoadSeg(childName);
    if ChildSegment ~= 0 then
	StartSeg*.ss_nextSeg := ChildSegment;
	StartSeg*.ss_jmpAddr := (ChildSegment << 2) + 4;
	ChildPort :=
	    CreateProc("ddb child", 0, (pretend(StartSeg, Segment_t) + 4) >> 2,
		       20000);
	if ChildPort = nil then
	    ignore UnLoadSeg(ChildSegment);
	    false
	else
	    true
	fi
    else
	false
    fi
corp;

/*
 * childUnLoad - unload a loaded child program.
 */

proc childUnLoad()void:

    ignore UnLoadSeg(ChildSegment);
corp;

/*
 * childRun - run the child program.
 */

proc childRun()bool:
    *Message_t message;

    if ChildHasRun then
	ChildPort :=
	    CreateProc("ddb child", 0, (pretend(StartSeg, Segment_t) + 4) >> 2,
		       20000);
	if ChildPort = nil then
	    return false;
	fi;
    fi;
    ChildHasRun := true;
    ignore WaitPort(ParentPort);
    message := GetMsg(ParentPort);
    ChildProcess := pretend(ChildPort - sizeof(Task_t), *Process_t);
    OldReturnAddress := pretend(ChildProcess*.pr_ReturnAddr - 4, *ulong)*;
    pretend(ChildProcess*.pr_ReturnAddr - 4, *ulong)* :=
	pretend(childTerminate, ulong);
    ChildProcess*.pr_COS := Open("*", MODE_NEWFILE);
    ReplyMsg(message);
    true
corp;

/*
 * childWait - wait for the running child to terminate.
 */

proc childWait()void:

    ParentWaiting := true;
    ignore WaitPort(ParentPort);
    ignore GetMsg(ParentPort);
    ParentWaiting := false;
corp;
