#ddb.g

char VERSION_1 = '0', VERSION_2 = '1';

uint
    INPUT_MAX = 256,
    PROGRAM_NAME_MAX = 256,
    NAME_MAX = 8;

type
    Name_t = [NAME_MAX] char;

uint
    COMMAND_COUNT = 7,

    CMD_UNKNOWN = 0,
    CMD_QUIT = 1,
    CMD_HELP = 2,
    CMD_SUSPEND = 3,
    CMD_LOAD = 4,
    CMD_UNLOAD = 5,
    CMD_RUN = 6,
    CMD_WAIT = 7;

[COMMAND_COUNT + 1] Name_t COMMAND_TABLE = (
    "quit",
    "help",
    "suspend",
    "load",
    "unload",
    "run",
    "wait",
    ""
);

[INPUT_MAX] char InputLine;
uint InputPos;
[PROGRAM_NAME_MAX] char ProgramName;
bool ProgramLoaded, ProgramRunning;

/*
 * skipBlanks - skip past blanks in the input line.
 */

proc skipBlanks()void:

    while InputLine[InputPos] = ' ' do
	InputPos := InputPos + 1;
    od;
corp;

/*
 * getWord - pick a "word" off of the input line.
 */

proc getWord(*char buffer; uint len)void:

    while InputLine[InputPos] ~= ' ' and InputLine[InputPos] ~= '\e' do
	if len > 1 then
	    buffer* := InputLine[InputPos];
	    buffer := buffer + sizeof(char);
	    len := len - 1;
	fi;
	InputPos := InputPos + 1;
    od;
    buffer* := '\e';
corp;

/*
 * lookupName - look up a name in a name table. Return '0' for not-found.
 */

proc lookupName(register *char name; *Name_t names)uint:
    register uint count;
    register *char p, q;

    count := 1;
    while
	if names*[0] = '\e' then
	    false
	else
	    p := &names*[0];
	    q := name;
	    while p* = q* and p* ~= '\e' do
		p := p + sizeof(char);
		q := q + sizeof(char);
	    od;
	    p* ~= q*
	fi
    do
	names := names + sizeof(Name_t);
	count := count + 1;
    od;
    if names*[0] = '\e' then
	0
    else
	count
    fi
corp;

/*
 * cmdHelp - the 'help' command.
 */

proc cmdHelp()void:

    putStr(
	"DDB commands are:\n"
	"  help - this help\n"
	"  quit - exit DDB\n"
	"  suspend - suspend DDB\n"
	"  load <program> - load a child program\n"
	"  unload - unload a loaded program\n"
	"  run - run the loaded program\n"
	"  wait - wait for the child program to finish\n"
    );
corp;

/*
 * cmdSuspend - the 'suspend' command.
 */

proc cmdSuspend()void:

    setSuspend(true);
corp;

/*
 * err - simple error message.
 */

proc err(*char m)void:

    putStr("*** ");
    putStr(m);
    putStr(". ***\n");
corp;

/*
 * progError - error message withe current program name in the middle.
 */

proc progError(*char m1, m2)void:

    putStr("*** ");
    putStr(m1);
    putStr(" '");
    putStr(&ProgramName[0]);
    putStr("' ");
    putStr(m2);
    putStr(". ***\n");
corp;

/*
 * tryToLoad - try to load the named program.
 */

proc tryToLoad()void:

    if childLoad(&ProgramName[0]) then
	ProgramLoaded := true;
	putStr("Program '");
	putStr(&ProgramName[0]);
	putStr("' loaded.\n");
    else
	progError("cannot load", "");
    fi;
corp;

/*
 * cmdLoad - the 'load' command.
 */

proc cmdLoad()void:

    if ProgramLoaded then
	progError("program", "is already loaded");
    else
	getWord(&ProgramName[0], PROGRAM_NAME_MAX);
	if ProgramName[0] = '\e' then
	    err("no program name given");
	else
	    tryToLoad();
	fi;
    fi;
corp;

/*
 * cmdUnLoad - the 'unload' command.
 */

proc cmdUnLoad()void:

    if ProgramRunning then
	progError("program", "is still running");
    elif ProgramLoaded then
	childUnLoad();
	ProgramLoaded := false;
    else
	err("there is no program loaded");
    fi;
corp;

/*
 * cmdRun - the 'run' command.
 */

proc cmdRun()void:

    if ProgramRunning then
	progError("program", "is already running");
    elif not ProgramLoaded then
	err("there is no program loaded");
    else
	putStr(">>> Starting child program '");
	putStr(&ProgramName[0]);
	putStr("' <<<\n");
	putStr(">>> Suspending <<<\n");
	setSuspend(true);
	if childRun() then
	    ProgramRunning := true;
	else
	    progError("CreateProc cannot rerun program", "(unsuspending)");
	fi;
    fi;
corp;

/*
 * cmdWait - the 'wait' command.
 */

proc cmdWait()void:

    if ProgramRunning then
	childWait();
	ProgramRunning := false;
    else
	err("there is no program running");
    fi;
corp;

/*
 * childDone - called by 'child' code if child terminates when we aren't
 *	waiting for it.
 *	NOTE: this routine is called from the child process!
 */

proc childDone()void:

    startSpecial();
    putStr("\n>>> Child program '");
    putStr(&ProgramName[0]);
    putStr("' terminated. <<<\n\n");
    endSpecial();
    ProgramRunning := false;
corp;

/*
 * commands - command processor
 */

proc commands()void:
    extern GetPars(*ulong pLen; **char pPtr)void;
    [NAME_MAX] char cmdName;
    *char xptr;
    ulong xlen;
    register *char p, q;
    register ulong len;
    uint cmd;

    GetPars(&xlen, &xptr);
    p := xptr;
    len := xlen;
    while len ~= 0 and (p* = ' ' or p* = '\t') do
	p := p + sizeof(char);
	len := len - 1;
    od;
    if len = 0 or p* = '\e' or p* = '\r' or p* = '\n' then
	ProgramName[0] := '\e';
    else
	q := &ProgramName[0];
	while len ~= 0 and p* ~= ' ' and p* ~= '\e' and
	    p* ~= '\r' and p* ~= '\n'
	do
	    q* := p*;
	    q := q + sizeof(char);
	    p := p + sizeof(char);
	    len := len - 1;
	od;
	q* := '\e';
    fi;

    putStr("DDB V\(VERSION_1).\(VERSION_2) by Chris Gray\n\n");
    putChar('\n');
    ProgramLoaded := false;
    ProgramRunning := false;
    if ProgramName[0] ~= '\e' then
	tryToLoad();
    fi;

    while
	readLine("-> ", &InputLine[0], INPUT_MAX);
	InputPos := 0;
	skipBlanks();
	if InputLine[InputPos] = '\e' then
	    cmd := CMD_QUIT;	/* kludge! */
	    true
	else
	    getWord(&cmdName[0], NAME_MAX);
	    cmd := lookupName(&cmdName[0], &COMMAND_TABLE[0]);
	    cmd ~= CMD_QUIT
	fi
    do
	skipBlanks();
	case cmd
	incase CMD_UNKNOWN:
	    putStr("*** unknown command '");
	    putStr(&cmdName[0]);
	    putStr("', use 'help' for help.\n");
	incase CMD_HELP:
	    cmdHelp();
	incase CMD_SUSPEND:
	    cmdSuspend();
	incase CMD_LOAD:
	    cmdLoad();
	incase CMD_UNLOAD:
	    cmdUnLoad();
	incase CMD_RUN:
	    cmdRun();
	incase CMD_WAIT:
	    cmdWait();
	esac;
    od;
    if ProgramRunning then
	putStr("Waiting for program '");
	putStr(&ProgramName[0]);
	putStr("' to complete.\n");
	childWait();
    fi;
    if ProgramLoaded then
	putStr("Unloading program '");
	putStr(&ProgramName[0]);
	putStr("'.\n");
	childUnLoad();
    fi;
corp;
