#drinc:exec/miscellaneous.g
#drinc:exec/nodes.g
#drinc:exec/ports.g
#drinc:exec/tasks.g
#drinc:exec/memory.g
#drinc:libraries/dos.g
#drinc:libraries/dosextens.g

*Message_t Message;
*MsgPort_t ParentPort, ChildPort;
ulong OldReturnAddress;
bool ChildPortOK;

/*
 * printStr - print a simple message to standard output.
 */

proc printStr(register *char message)void:
    register *char p;

    p := message;
    while p* ~= '\e' do
	p := p + sizeof(char);
    od;
    ignore Write(Output(), message, p - message);
corp;

/*
 * printNum - print an unsigned 32 bit number to standard output.
 */

proc printNum(register ulong n)void:
    [12] char buffer;
    register *char p;

    p := &buffer[11];
    p* := '\e';
    while
	p := p - sizeof(char);
	p* := n % 10 + '0';
	n := n / 10;
	n ~= 0
    do
    od;
    printStr(p);
corp;

/*
 * abort - print a message and Exit.
 */

proc abort(*char message)void:

    printStr(message);
    Exit(RETURN_FAIL);
corp;

/*
 * childStartup - routine inserted into the beginning of the child. It creates
 *	a port which it can receive messages, on then sends a message to the
 *	parent. This allows the parent to wait for child startup. The child
 *	then waits for a reply from the parent. This allows the parent to hold
 *	the child until it is ready to release it.
 */

proc childStartup()void:
    uint
	OP_MOVEL = 0x2000,
	M_DDIR = 0o0,
	M_ADIR = 0o1,
	M_INC = 0o3,
	M_DEC = 0o4,
	R_D1 = 1,
	R_A1 = 1,
	R_A6 = 6,
	R_SP = 7;

    code(
	OP_MOVEL | R_SP << 9 | M_DEC << 6 | M_DDIR << 3 | R_D1,
	OP_MOVEL | R_SP << 9 | M_DEC << 6 | M_ADIR << 3 | R_A1,
	OP_MOVEL | R_SP << 9 | M_DEC << 6 | M_ADIR << 3 | R_A6
    );
    ChildPort := CreatePort(nil, 0);
    Message*.mn_Node.ln_Type := NT_MESSAGE;
    Message*.mn_ReplyPort := ChildPort;
    PutMsg(ParentPort, Message);
    if ChildPort = nil then
	ChildPortOK := false;
	Exit(RETURN_FAIL);
    else
	ChildPortOK := true;
	ignore WaitPort(ChildPort);
	ignore GetMsg(ChildPort);
    fi;
    code(
	OP_MOVEL | R_A6 << 9 | M_ADIR << 6 | M_INC << 3 | R_SP,
	OP_MOVEL | R_A1 << 9 | M_ADIR << 6 | M_INC << 3 | R_SP,
	OP_MOVEL | R_D1 << 9 | M_DDIR << 6 | M_INC << 3 | R_SP
    );
corp;

/*
 * childTerminate - routine inserted into the child's termination. This
 *	does another synchronization with the parent, then jumps indirect
 *	through the original DOS return address to continue with the Exit.
 */

proc childTerminate()void:
    uint
	OP_MOVEL = 0x2000,
	OP_JMP = 0x4ec0,
	M_ADIR = 1,
	M_INDIR = 2,
	M_SPECIAL = 0o7,
	M_ABS_LONG = 0o1,
	A0 = 0;

    Message*.mn_Node.ln_Type := NT_MESSAGE;
    Message*.mn_ReplyPort := ChildPort;
    PutMsg(ParentPort, Message);
    ignore WaitPort(ChildPort);
    ignore GetMsg(ChildPort);
    FreeMem(Message, sizeof(Message_t));
    DeletePort(ChildPort);
    code(
	OP_MOVEL | A0 << 9 | M_ADIR << 6 | M_SPECIAL << 3 | M_ABS_LONG,
	OldReturnAddress,
	OP_JMP | M_INDIR << 3 | A0
    );
corp;

/*
 * main - parse parameters, load program, start it, calculate time, print it.
 */

proc main()void:
    uint
	OP_MOVEL = 0x2000,
	OP_JSR = 0x4e80,
	OP_JMP = 0x4ec0,
	M_DDIR = 0o0,
	M_ADIR = 0o1,
	M_SPECIAL = 0o7,
	M_ABS_LONG = 0o1,
	ABS_LONG = M_SPECIAL << 3 | M_ABS_LONG,
	R_A0 = 0,
	R_D0 = 0;
    type startSeg_t = struct {
	ulong ss_length;
	ulong ss_nextSeg;
	uint ss_jsrCode;
	ulong ss_jsrAddr;
	uint ss_moveAddrCode;
	ulong ss_moveAddrAddr;
	uint ss_moveLenCode;
	ulong ss_moveLenAddr;
	uint ss_jmpCode;
	ulong ss_jmpAddr;
    };
    extern GetPars(*ulong pLen; **char pPtr)void;
    DateStamp_t startTime, endTime;
    Segment_t childSegment;
    *startSeg_t startSeg;
    *MsgPort_t childProcPort;
    *Process_t childProcess;
    *char ptr, start;
    ulong len;
    long minutes, ticks, fraction;

    if OpenExecLibrary(0) ~= nil then
	if OpenDosLibrary(0) ~= nil then

	    /* parse command line - extract program name */

	    GetPars(&len, &ptr);
	    while len ~= 0 and (ptr* = ' ' or ptr* = '\t') do
		ptr := ptr + sizeof(char);
		len := len - 1;
	    od;
	    if len = 0 or ptr* = '\e' or ptr* = '\r' or ptr* = '\n' then
		abort("Use is: time prog [args]\n");
	    fi;
	    start := ptr;
	    while len ~= 0 and ptr* ~= ' ' and ptr* ~= '\e' and
		ptr* ~= '\r' and ptr* ~= '\n'
	    do
		ptr := ptr + sizeof(char);
		len := len - 1;
	    od;
	    if ptr* ~= '\e' then
		ptr* := '\e';
		ptr := ptr + sizeof(char);
		len := len - 1;
	    fi;

	    /* load that program */

	    childSegment := LoadSeg(start);
	    if childSegment = 0 then
		printStr("time: could not load '");
		printStr(start);
		abort("'\n");
	    fi;

	    /* create a new fake segment to become the first segment of the
	       loaded program. This allows us to get control after the child
	       task is fully created, but before it really starts running. */

	    startSeg := AllocMem(sizeof(startSeg_t), MEMF_PUBLIC);
	    if startSeg = nil then
		ignore UnLoadSeg(childSegment);
		abort("time: could not allocate start segment\n");
	    fi;
	    startSeg*.ss_length := sizeof(startSeg_t);
	    startSeg*.ss_nextSeg := childSegment;
	    startSeg*.ss_jsrCode := OP_JSR | M_SPECIAL << 3 | M_ABS_LONG;
	    startSeg*.ss_jsrAddr := pretend(childStartup, ulong);
	    startSeg*.ss_moveAddrCode :=
		OP_MOVEL | R_A0 << 9 | M_ADIR << 6 | ABS_LONG;
	    startSeg*.ss_moveAddrAddr := pretend(&ptr, ulong);
	    startSeg*.ss_moveLenCode :=
		OP_MOVEL | R_D0 << 9 | M_DDIR << 6 | ABS_LONG;
	    startSeg*.ss_moveLenAddr := pretend(&len, ulong);
	    startSeg*.ss_jmpCode := OP_JMP | M_SPECIAL << 3 | M_ABS_LONG;
	    startSeg*.ss_jmpAddr := (childSegment << 2) + 4;

	    /* create the first of the two synchronization ports */

	    ParentPort := CreatePort(nil, 0);
	    if ParentPort = nil then
		FreeMem(startSeg, sizeof(startSeg_t));
		ignore UnLoadSeg(childSegment);
		abort("time: could not create main port\n");
	    fi;

	    Message := AllocMem(sizeof(Message_t), MEMF_PUBLIC | MEMF_CLEAR);
	    if Message = nil then
		DeletePort(ParentPort);
		FreeMem(startSeg, sizeof(startSeg_t));
		ignore UnLoadSeg(childSegment);
		abort("time: could not create message\n");
	    fi;

	    /* create the child process from the constructed segment list.
	       NOTE: it starts running immediately! */

	    childProcPort := CreateProc("ddb child", 0,
			   (pretend(startSeg, Segment_t) + 4) >> 2, 4000);
	    if childProcPort = nil then
		FreeMem(Message, sizeof(Message_t));
		DeletePort(ParentPort);
		FreeMem(startSeg, sizeof(startSeg_t));
		ignore UnLoadSeg(childSegment);
		abort("time: could not create child process\n");
	    fi;

	    /* synchronize - wait for child to send the message to us */

	    ignore WaitPort(ParentPort);
	    ignore GetMsg(ParentPort);

	    /* we check to see if the child was able to create ChildPort, and
	       if not, don't continue any further */

	    if not ChildPortOK then
		FreeMem(Message, sizeof(Message_t));
		DeletePort(ParentPort);
/*
		FreeMem(startSeg, sizeof(startSeg_t));
		ignore UnLoadSeg(childSegment);
*/
		abort("time: child could not create reply port\n");
	    fi;

	    /* fix up its return address to call our new return routine */

	    childProcess :=
		pretend(childProcPort - sizeof(Task_t), *Process_t);
	    OldReturnAddress :=
		pretend(childProcess*.pr_ReturnAddr - 4, *ulong)*;
	    pretend(childProcess*.pr_ReturnAddr - 4, *ulong)* :=
		pretend(childTerminate, ulong);
	    childProcess*.pr_COS := Open("*", MODE_NEWFILE);

	    /* get the start time, and allow the child to execute */

	    DateStamp(&startTime);
	    ReplyMsg(Message);

	    /* child process runs */

	    /* synchronize - wait for child to send terminate message to us */

	    ignore WaitPort(ParentPort);
	    ignore GetMsg(ParentPort);

	    /* get the end time and allow the child to go away */

	    DateStamp(&endTime);
	    ReplyMsg(Message);

	    /* free up all our dynamic stuff */

	    DeletePort(ParentPort);
	    FreeMem(startSeg, sizeof(startSeg_t));
	    if not UnLoadSeg(childSegment) then
		abort("time: can't unload program\n");
	    fi;

	    /* figure out and print the execution time */

	    minutes := endTime.ds_Minute - startTime.ds_Minute;
	    ticks := endTime.ds_Tick - startTime.ds_Tick;
	    if ticks < 0 then
		ticks := ticks + 50 * 60;
		minutes := minutes - 1;
	    fi;
	    fraction := ticks % 50 * 2;
	    printStr("Run time = ");
	    if minutes = 0 then
		printNum(ticks / 50);
		printStr(".");
		if fraction < 10 then
		    printStr("0");
		fi;
		printNum(fraction);
		printStr(" seconds\n");
	    else
		printNum(minutes);
		printStr(" minutes, ");
		printNum(ticks / 50);
		printStr(".");
		if fraction < 10 then
		    printStr("0");
		fi;
		printNum(fraction);
		printStr(" seconds (");
		minutes := minutes * (50 * 60);
		ticks := ticks + minutes;
		printNum(ticks / 50);
		printStr(".");
		fraction := ticks % 50 * 2;
		if fraction < 10 then
		    printStr("0");
		fi;
		printNum(fraction);
		printStr(" seconds)\n");
	    fi;

	    CloseDosLibrary();
	fi;
	CloseExecLibrary();
    fi;
corp;
