#drinc:exec/miscellaneous.g
#drinc:exec/nodes.g
#drinc:exec/ports.g
#drinc:exec/tasks.g
#drinc:exec/memory.g
#drinc:libraries/dos.g
#drinc:libraries/dosextens.g

#system.g

*char PORT_NAME = "system portX";
uint
    PORT_NAME_LENGTH = 13,
    MAX_TRIES = 26,
    REDIR_NAME_LENGTH = 256;

    uint
	OP_MOVEL = 0x2000,
	OP_JMP = 0x4ec0,
	OP_JSR = 0x4e80,
	M_DDIR = 0o0,
	M_ADIR = 0o1,
	M_INDIR = 0o2,
	M_SPECIAL = 0o7,
	M_ABS_LONG = 0o1,
	A0 = 0,
	D0 = 0;

type
    StartSeg_t = struct {
	ulong ss_length;
	ulong ss_nextSeg;
	uint ss_jsrCode;
	ulong ss_jsrAddr;
	uint ss_jmpCode;
	ulong ss_jmpAddr;
    };

[PORT_NAME_LENGTH] char PortName;
*MsgPort_t ParentPort;
*Message_t TheMessage;
*Process_t ChildProcess;
*char ParPtr;
[REDIR_NAME_LENGTH] char OutputName, InputName;
ulong OldReturnAddress, ParLen;
bool CanWait;

/*
 * childStartup - routine inserted into the beginning of the child
 */

proc childStartup()void:
    register *MsgPort_t replyPort;

    replyPort := CreatePort(nil, 0);
    if replyPort ~= nil then
	CanWait := true;
	TheMessage*.mn_ReplyPort := replyPort;
	PutMsg(ParentPort, TheMessage);
	ignore WaitPort(replyPort);
	ignore GetMsg(replyPort);
	DeletePort(replyPort);
    else
	CanWait := false;
	PutMsg(ParentPort, TheMessage);
    fi;
    code(
	OP_MOVEL | D0 << 9 | M_DDIR << 6 | M_SPECIAL << 3 | M_ABS_LONG,
	ParLen,
	OP_MOVEL | A0 << 9 | M_ADIR << 6 | M_SPECIAL << 3 | M_ABS_LONG,
	ParPtr
    );
corp;

/*
 * childTerminate - routine inserted into the child's termination.
 */

proc childTerminate()void:

    PutMsg(ParentPort, TheMessage);
    code(
	OP_MOVEL | A0 << 9 | M_ADIR << 6 | M_SPECIAL << 3 | M_ABS_LONG,
	OldReturnAddress,
	OP_JMP | M_INDIR << 3 | A0
    );
corp;

proc runIt(*char commandName; *StartSeg_t startSeg; ulong stackSize)ulong:
    register Segment_t childSegment;
    register ulong errorCode;
    *ulong retAddrPtr;
    *MsgPort_t childPort;

    errorCode := SYS_OK;
    childSegment := LoadSeg(commandName);
    if childSegment ~= 0 then
	startSeg*.ss_nextSeg := childSegment;
	startSeg*.ss_jmpAddr := (childSegment << 2) + 4;
	childPort :=
	    CreateProc("system child", 0,
		       (pretend(startSeg, Segment_t) + 4) >> 2, stackSize);
	if childPort ~= nil then
	    ChildProcess*.pr_COS := Open(&OutputName[0], MODE_NEWFILE);
	    ChildProcess*.pr_CIS := Open(&InputName[0], MODE_OLDFILE);
	    ignore WaitPort(ParentPort);
	    ignore GetMsg(ParentPort);
	    if CanWait then
		ChildProcess := pretend(childPort - sizeof(Task_t),*Process_t);
		retAddrPtr := pretend(ChildProcess*.pr_ReturnAddr - 4, *ulong);
		OldReturnAddress := retAddrPtr*;
		retAddrPtr* := pretend(childTerminate, ulong);
		ReplyMsg(TheMessage);
		ignore WaitPort(ParentPort);
		ignore GetMsg(ParentPort);
		if not UnLoadSeg(childSegment) then
		    errorCode := SYS_NO_UNLOAD;
		fi;
		Close(ChildProcess*.pr_CIS);
		Close(ChildProcess*.pr_COS);
	    else
		errorCode := SYS_NO_WAIT;
	    fi;
	else
	    ignore UnLoadSeg(childSegment);
	    errorCode := SYS_NO_PROC;
	fi;
    else
	errorCode := SYS_NO_LOAD;
    fi;
    errorCode
corp;

proc copy(register *char a, b)void:

    while b* ~= '\e' do
	a* := b*;
	a := a + sizeof(char);
	b := b + sizeof(char);
    od;
corp;

proc findStart(register *char p)*char:

    while p* = ' ' or p* = '\t' do
	p := p + sizeof(char);
    od;
    p
corp;

proc findEnd(register *char p)*char:

    while p* ~= '\e' and p* ~= ' ' and p* ~= '\t' and
	p* ~= '\n' and p* ~= '\r'
    do
	p := p + sizeof(char);
    od;
    p
corp;

/*
 * system - run a given program, then wait for it to complete.
 */

proc system(register *char cmd; ulong stackSize)ulong:
    register *char p, q, r;
    register *StartSeg_t startSeg @ r;
    register ulong errorCode;
    register ushort tries;
    register char oldCh;
    bool hadOutput, hadInput;

    if stackSize < 4000 then
	stackSize := 4000;
    fi;
    copy(&InputName[0], "NIL:");
    copy(&OutputName[0], "NIL:");
    cmd := findStart(cmd);
    p := findEnd(cmd);
    if p* = '\e' then
	ParPtr := p;
	ParLen := 0;
    else
	p* := '\e';
	p := p + sizeof(char);
	p := findStart(p);
	hadOutput := false;
	hadInput := false;
	while p* = '>' and not hadOutput or p* = '<' and not hadInput do
	    q := p;
	    p := p + sizeof(char);
	    p := findStart(p);
	    r := p;
	    p := findEnd(p);
	    oldCh := p*;
	    p* := '\e';
	    if q* = '>' then
		copy(&OutputName[0], r);
		hadOutput := true;
	    else
		copy(&InputName[0], r);
		hadInput := true;
	    fi;
	    if oldCh = '\e' then
		q* := '\e';
	    else
		p := p + sizeof(char);
		copy(q, p);
		p := q;
	    fi;
	od;
	ParPtr := p;
	while p* ~= '\e' do
	    p := p + sizeof(char);
	od;
	ParLen := (p - ParPtr) / sizeof(char);
    fi;
    errorCode := SYS_OK;
    startSeg := AllocMem(sizeof(StartSeg_t), MEMF_PUBLIC);
    if startSeg ~= nil then
	startSeg*.ss_length := sizeof(StartSeg_t);
	startSeg*.ss_jsrCode := OP_JSR | M_SPECIAL << 3 | M_ABS_LONG;
	startSeg*.ss_jsrAddr := pretend(childStartup, ulong);
	startSeg*.ss_jmpCode := OP_JMP | M_SPECIAL << 3 | M_ABS_LONG;
	copy(&PortName[0], PORT_NAME);
	tries := 0;
	while
	    PortName[PORT_NAME_LENGTH - 2] := 'a' + tries;
	    FindPort(&PortName[0]) ~= nil and tries ~= MAX_TRIES
	do
	    tries := tries + 1;
	od;
	if tries ~= MAX_TRIES then
	    ParentPort := CreatePort(&PortName[0], 0);
	    if ParentPort ~= nil then
		TheMessage :=
		    AllocMem(sizeof(Message_t), MEMF_PUBLIC | MEMF_CLEAR);
		if TheMessage ~= nil then
		    TheMessage*.mn_Node.ln_Type := NT_MESSAGE;
		    errorCode := runIt(cmd, startSeg, stackSize);
		    FreeMem(TheMessage, sizeof(Message_t));
		else
		    errorCode := SYS_NO_MEM;
		fi;
		DeletePort(ParentPort);
	    else
		errorCode := SYS_NO_PORT;
	    fi;
	else
	    errorCode := SYS_NO_PORT;
	fi;
	if errorCode ~= SYS_NO_WAIT then
	    FreeMem(startSeg, sizeof(StartSeg_t));
	fi;
    else
	errorCode := SYS_NO_MEM;
    fi;
    errorCode
corp;
